/*
 * BSS Config related declarations and exported functions for
 * Broadcom 802.11abg Networking Device Driver
 *
 * Copyright 2005-2006, Broadcom Corporation
 * All Rights Reserved.                
 *                                     
 * THIS SOFTWARE IS OFFERED "AS IS", AND BROADCOM GRANTS NO WARRANTIES OF ANY
 * KIND, EXPRESS OR IMPLIED, BY STATUTE, COMMUNICATION OR OTHERWISE. BROADCOM
 * SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A SPECIFIC PURPOSE OR NONINFRINGEMENT CONCERNING THIS SOFTWARE.
 * $Id$
 */
#ifndef _WLC_BSSCFG_H
#define _WLC_BSSCFG_H


struct wlc_bsscfg;
struct wl_if;

/* virtual interface */
struct wlc_if {
	int		type;		/* WLC_IFTYPE_BSS */
	struct wl_if	*wlif;		/* pointer to wlif */
	union {
		struct ether_addr wds_ea;	/* wds key */
		struct wlc_bsscfg *bsscfg;	/* pointer to bss config */
	} u;
};

/* virtual interface types */
#define WLC_IFTYPE_BSS 1		/* virt interface for a bsscfg */

/* BSS configuration state */
struct wlc_bsscfg {
	bool		up;		/* is this configuration up */
	bool		enable;		/* is this configuration enabled */
	bool		_ap;		/* is this configuration an AP */
	struct wlc_if	wlcif;		/* virtual interface, NULL for primary bsscfg */
	uint8		SSID_len;	/* the length of SSID */
	uint8		SSID[32];	/* SSID string */
	bool		closednet;	/* hide ssid info */
	bool		ap_isolate;	/* true if isolating associated STA devices */
	int		prb_resp_len;	/* probe response length */
	uint8		*prb_resp;	/* probe response, 802.11 hdr to end, not including FCS */

#ifdef PER_BSS_MACLIST
	uint		nmac;		/* # of entries on maclist array */
	int		macmode;	/* allow/deny stations on maclist array */
	struct ether_addr maclist[MAXMACLIST];	/* list of source MAC addrs to match */
#endif	/* PER_BSS_MACLIST */

	/* security */
	uint32		wsec;		/* wireless security bitvec */
	uint32		pwsec;		/* Pairwise wireless security bitvec */
	uint32		gwsec;		/* Group wireless security bitvec */
	int		auth;		/* 802.11 authentication: Open, Shared Key, WPA */
	bool		wsec_restrict;	/* drop unencrypted packets if wsec is enabled */
	bool		eap_restrict;	/* restrict data until 802.1X auth succeeds */
	bool 		rx_unencrypted_eapol; /* accept unencrypted eapol frames */
	uint8		WPA_auth;	/* WPA: authenticated key management */
	wsec_iv_t	wpa_none_txiv;	/* global txiv for WPA_NONE, tkip and aes */
	int		wsec_index;	/* 0-3: default tx key, -1: not set */
	wsec_key_t	*bss_def_keys[WLC_DEFAULT_KEYS];	/* default key storage */
	/* TKIP countermeasures */
	bool		tkip_countermeasures;	/* flags TKIP no-assoc period */
	uint32		tk_cm_dt;	/* detect timer */
	uint32		tk_cm_bt;	/* blocking timer */
	bool		tk_cm_activate;	/* activate countermeasures after EAPOL-Key sent */
};

/* forward declarations */
typedef struct wlc_bsscfg wlc_bsscfg_t;
typedef struct wlc_if wlc_if_t;
struct wlc_info;

void wlc_bsscfg_init(struct wlc_info *wlc, wlc_bsscfg_t *bsscfg);
wlc_bsscfg_t* wlc_bsscfg_get(struct wlc_info *wlc, int idx, int *perr);
void wlc_bsscfg_free(struct wlc_info *wlc, wlc_bsscfg_t *bsscfg);
void wlc_bsscfg_disablemulti(struct wlc_info *wlc);
void wlc_bsscfg_disable(struct wlc_info *wlc, wlc_bsscfg_t *cfg);
void wlc_bsscfg_down(struct wlc_info *wlc, wlc_bsscfg_t *cfg);
int wlc_bsscfg_up(struct wlc_info *wlc, wlc_bsscfg_t *cfg);
int wlc_bsscfg_enable(struct wlc_info *wlc, wlc_bsscfg_t *cfg);
wlc_bsscfg_t* wlc_bsscfg_lookup(struct wlc_info *wlc, uchar* ssid, uint ssid_len);
wlc_bsscfg_t* wlc_bsscfg_find(struct wlc_info *wlc, int idx, int *perr);
bool wlc_bsscfg_notxc(struct wlc_info *wlc);


#define WLC_GET_BSSCFG(wl, idx) \
	(((idx) < WLC_MAXBSSCFG && (idx) >= 0) ? ((wl)->bsscfg[idx]) : 0)
#define WLC_BSSCFG_AUTH(cfg) ((cfg)->auth)

#define WLC_BSSCFG_WEPKEY(cfg, idx) \
	(((idx) < WLC_DEFAULT_KEYS && (int)(idx) >= 0) ? \
	((cfg)->bss_def_keys[idx]) : 0)

#endif	/* _WLC_BSSCFG_H */
