/*
 * Required functions exported by the wlc_phy.c (phy-dependent)
 * to common (os-independent) driver code.
 *
 * Copyright 2005-2006, Broadcom Corporation
 * All Rights Reserved.                
 *                                     
 * THIS SOFTWARE IS OFFERED "AS IS", AND BROADCOM GRANTS NO WARRANTIES OF ANY
 * KIND, EXPRESS OR IMPLIED, BY STATUTE, COMMUNICATION OR OTHERWISE. BROADCOM
 * SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A SPECIFIC PURPOSE OR NONINFRINGEMENT CONCERNING THIS SOFTWARE.
 *
 * $Id$
 */

#ifndef _wlc_phy_h_
#define _wlc_phy_h_



/* Public phy info structure */
struct phy_pub {
	uint		phy_type;		/* PHY_TYPE_XX */
	uint		phy_rev;		/* phy revision */
	uint		coreflags;		/* sbtml core/phy specific flags */
	uint		ana_rev;		/* analog core revision */
	uint		radio_channel;		/* current radio channel, A or B channel values */
	uint16		radio_code;		/* radio code */
	uint16		radioid;		/* radio id */
	uint8		radiorev;		/* radio revision */
	bool		hwpwrctrl;		/* ucode controls txpower instead of driver */
	bool		encore;			/* true if chipset is encore enabled */
	uint8		tx_power_max;		/* max power for all rates on this channel. Power
						 * offsets use this.
						 */
};

/* prototypes */
extern void *wlc_phy_attach(wlc_pub_t *pub, void *regs, int bandtype, void *wlc, void *phy_sh);
extern void wlc_phy_detach(wlc_phy_t *ppi);
extern void *wlc_phy_shared_attach(wlc_pub_t *pub);
extern void wlc_phy_shared_detach(wlc_pub_t *pub, void *phy_sh);
extern void wlc_phy_init(wlc_phy_t *ppi);
extern void wlc_phy_cal_init(wlc_phy_t *ppi);
extern int
wlc_phy_ioctl(wlc_phy_t *ppi, int cmd, int len, int *pval, bool bool_val, void *arg, bool *ta_ok);
extern void wlc_switch_radio(wlc_phy_t *ppi, bool on);
extern void wlc_clear_tssi(wlc_phy_t *ppi);
extern uint wlc_radiocode2channel(uint radiocode, uint phytype);
extern uint16 wlc_channel2radiocode(uint16 channel, uint phytype);
extern void wlc_band_channels(wlc_phy_t *ppi, uint band, chanvec_t *channels);
extern uint8 wlc_band_first_channel(wlc_phy_t *ppi, uint band);
extern void wlc_phy_set_channel(wlc_phy_t *ppi, uint channel);
extern void wlc_phy_freqtrack_start(wlc_phy_t *ppi);
extern void wlc_phy_freqtrack_end(wlc_phy_t *ppi);

extern int wlc_update_rssi_ma(wlc_phy_t *ppi, d11rxhdr_t *rxh);
extern int wlc_compute_rssi(wlc_phy_t *ppi, d11rxhdr_t *rxh);
extern int8 wlc_compute_phy_noise(wlc_phy_t *ppi);

/* phy/radio reg access */
extern void wlc_phyreg_enter(wlc_phy_t *ppi);
extern void wlc_phyreg_exit(wlc_phy_t *ppi);
extern void wlc_radioreg_enter(wlc_phy_t *ppi);
extern void wlc_radioreg_exit(wlc_phy_t *ppi);

extern void wlc_phy_BSSinit(wlc_phy_t *ppi, int rssi);

/* txpwr */
extern void wlc_phy_11h_pwr(wlc_phy_t *ppi, dot11_power_cnst_t *local_pwr);
extern void wlc_txpwr_limits(wlc_phy_t *ppi, uint channel, int *min, int *max, int txp_rate_idx);
extern void wlc_adjust_txpwr_limit(wlc_phy_t *ppi, uint channel);
extern void wlc_txpower_recalc_target(wlc_phy_t *ppi, int channel, uint8 *chan_txpwr);
extern void wlc_phy_update_txpwr_shmem(wlc_phy_t *ppi, bool war);

extern bool wlc_phy_radio_disabled(void *regs);
extern void wlc_phy_analogon(void *regs);
extern void wlc_phy_analogoff(void *regs);




#endif	/* _wlc_phy_h_ */
