/*
 * HND PIO module
 * Broadcom 802.11abg Networking Device Driver
 *
 * Copyright 2005-2006, Broadcom Corporation
 * All Rights Reserved.                
 *                                     
 * THIS SOFTWARE IS OFFERED "AS IS", AND BROADCOM GRANTS NO WARRANTIES OF ANY
 * KIND, EXPRESS OR IMPLIED, BY STATUTE, COMMUNICATION OR OTHERWISE. BROADCOM
 * SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A SPECIFIC PURPOSE OR NONINFRINGEMENT CONCERNING THIS SOFTWARE.
 *
 * $Id$
 */
#ifndef WLPIO
#error "WLPIO is not defined"
#endif	/* WLPIO */

#include <wlc_cfg.h>
#include <typedefs.h>
#include <bcmdefs.h>
#include <osl.h>

#include <sbutils.h>
#include <bcmendian.h>
#include <bcmutils.h>
#include <wlioctl.h>

#include <sbhndpio.h>
#include <sbhnddma.h>
#include <hnddma.h>
#include <d11.h>
#include <wlc_rate.h>
#include <wlc_key.h>
#include <wlc_channel.h>
#include <wlc_pub.h>
#include <wlc_bsscfg.h>
#include <wlc_pio.h>
#include <wlc.h>


/* private pio structure */
typedef struct pio_info_s {
	pio_t		piopub;			/* public structure of pio module */
	wlc_info_t	*wlc;			/* common code handler */
	wlc_pub_t	*pub;			/* public common code handler */
	uint		_fifo;			/* fifo number */
	bool		pio4;			/* PIO 4 bytes or 2 bytes */
	pio2regs_t	*p2txregs;		/* pio2 trx regs */
	pio2regs_t	*p2rxregs;		/* pio2 rx regs */
	pio4regs_t	*p4txregs;		/* pio4 tx regs */
	pio4regs_t	*p4rxregs;		/* pio4 rx regs */
	/* pio */
	struct pktq	txpioq;			/* tx sw queue, wait txstatus to reclaim */
	uint		txfrmcnt;		/* tx frame count */
	uint		txdatacnt;		/* tx data count */
	uint		txfifolimit;		/* tx fifo depth */

} pio_info_t;

#define PIO_INFO(pio)	(pio_info_t *)pio

/* local prototype */

/* pio common routines for both pio2 and pio4 */
static int _wlc_pio_com_detach(pio_info_t *pt);
static int _wlc_pio_com_init(pio_info_t *pt);
static int _wlc_pio_com_reset(pio_info_t *pt);
static int _wlc_pio_com_txsuspend(pio_info_t *pt);
static bool _wlc_pio_com_txsuspended(pio_info_t *pt);
static int _wlc_pio_com_txresume(pio_info_t *pt);
static bool _wlc_pio_com_txavailable(pio_info_t *pt, uint len, uint nfrags);
static bool _wlc_pio_com_rxfrmrdy(pio_info_t *pt);
static int _wlc_pio_com_cntupd(pio_info_t *pt, uint len);
static char* _wlc_pio_com_dump(pio_info_t *pt, char *buf);
static void* _wlc_pio_getnexttxp(pio_info_t *pt);
static int   _wlc_pio_txreclaim(pio_info_t *pt);
static int   _wlc_pio_txfifodepthset(pio_info_t *pt, uint len);
static uint  _wlc_pio_txfifodepthget(pio_info_t *pt);

/* pio routines for 2 bytes pio */
static int _wlc_pio2_tx(pio_info_t *pt, void *p0);
static void _wlc_pio2_xmtfifo(pio_info_t *pioh, pio2regs_t *regs, uchar *va, uint len);
static void *_wlc_pio2_rx(pio_info_t *pt);

/* pio routines for 4 bytes pio */
static int _wlc_pio4_tx(pio_info_t *pt, void *p0);
static void _wlc_pio4_xmtfifo(pio_info_t *pt, pio4regs_t *regs, uchar *va, uint len);
static void *_wlc_pio4_rx(pio_info_t *pt);

static bool _wlc_pio_rxcheck(pio_info_t *pt, int len, void **p);

/* len is rounded up since free fifo length in ucode counts in multiples of 4 */
#define PIO_ADD_TX_CNT(pt, len) {	\
	pt->txfrmcnt++; \
	ASSERT(pt->txfrmcnt <= D11_MAX_TX_FRMS); \
	pt->txdatacnt += ROUNDUP(len, 4); \
	ASSERT(pt->txdatacnt <= pt->txfifolimit);	\
	}
#define PIO_SUB_TX_CNT(pt, len) {	\
	ASSERT(pt->txfrmcnt); \
	pt->txfrmcnt--;	\
	ASSERT(pt->txdatacnt >= ROUNDUP(len, 4));	\
	pt->txdatacnt -= ROUNDUP(len, 4);	\
	}

pio_t*
wlc_pio_attach(wlc_pub_t *pub, wlc_info_t *cwlc, uint fifo, void *pioregstx, void *pioregsrx)
{
	pio_info_t *pt;
	piof_t	pio_fns;

	if ((pt = (pio_info_t *)MALLOC(pub->osh, sizeof(pio_info_t))) == NULL) {
		WL_ERROR(("wlc_pio_attach: out of memory, malloced %d bytes", MALLOCED(pub->osh)));
		return NULL;
	}
	bzero((char *)pt, sizeof(pio_info_t));
	pt->wlc = cwlc;
	pt->pub = pub;
	pt->_fifo = fifo;

	/*
	 * register pio functions
	 * To add a new one, add the type and array element in wlc_pio.h,
	 *    add handler here and in wlc_pio_register_fn
	 */
	pio_fns.detach = (pio_detach_t)_wlc_pio_com_detach;
	pio_fns.reset = (pio_reset_t)_wlc_pio_com_reset;
	pio_fns.init = (pio_init_t)_wlc_pio_com_init;
	pio_fns.txsuspend = (pio_txsuspend_t)_wlc_pio_com_txsuspend;
	pio_fns.txsuspended = (pio_txsuspended_t)_wlc_pio_com_txsuspended;
	pio_fns.txresume = (pio_txresume_t)_wlc_pio_com_txresume;
	pio_fns.txavailable = (pio_txavailable_t)_wlc_pio_com_txavailable;
	pio_fns.rxfrmrdy = (pio_rxfrmrdy_t)_wlc_pio_com_rxfrmrdy;
	if (WLC_UPDATE_STATS(wlc)) {
		pio_fns.dump = (pio_dump_t)_wlc_pio_com_dump;
	}
	else {
		pio_fns.dump = NULL;
	}
	pio_fns.cntupd = (pio_cntupd_t)_wlc_pio_com_cntupd;
	pio_fns.nexttxp = (pio_getnexttxp_t)_wlc_pio_getnexttxp;
	pio_fns.txreclaim = (pio_txreclaim_t)_wlc_pio_txreclaim;
	pio_fns.txdepthget = (pio_txfifodepthget_t)_wlc_pio_txfifodepthget;
	pio_fns.txdepthset = (pio_txfifodepthset_t)_wlc_pio_txfifodepthset;

	/* init pio reg pointer */
	if (D11REV_LT(pt->pub->corerev, 8)) {
		pt->p2txregs = (pio2regs_t *)pioregstx;
		pt->p2rxregs = (pio2regs_t *)pioregsrx;
		pio_fns.rx = (pio_rx_t)_wlc_pio2_rx;
		pio_fns.tx = (pio_tx_t)_wlc_pio2_tx;
		pt->pio4 = 0;
	} else {
		pt->p4txregs = (pio4regs_t *)pioregstx;
		pt->p4rxregs = (pio4regs_t *)pioregsrx;
		pio_fns.rx = (pio_rx_t)_wlc_pio4_rx;
		pio_fns.tx = (pio_tx_t)_wlc_pio4_tx;
		pt->pio4 = 1;
	}

	wlc_pio_register_fn((pio_t *)pt, &pio_fns);

	pktq_init(&pt->txpioq, 1, PKTQ_LEN_DEFAULT);
	ASSERT(pktq_avail(&pt->txpioq) >= D11_MAX_TX_FRMS);

	return (pio_t*)pt;
}

/* register pio mode function pointers */
void
wlc_pio_register_fn(pio_t *pioh, piof_t *fn)
{
	pioh->pio_fn.detach = fn->detach;
	pioh->pio_fn.reset = fn->reset;
	pioh->pio_fn.init = fn->init;
	pioh->pio_fn.txsuspend = fn->txsuspend;
	pioh->pio_fn.txsuspended = fn->txsuspended;
	pioh->pio_fn.txresume = fn->txresume;
	pioh->pio_fn.rxfrmrdy = fn->rxfrmrdy;
	pioh->pio_fn.txavailable = fn->txavailable;
	pioh->pio_fn.dump = fn->dump;
	pioh->pio_fn.cntupd = fn->cntupd;
	pioh->pio_fn.nexttxp = fn->nexttxp;
	pioh->pio_fn.txreclaim = fn->txreclaim;
	pioh->pio_fn.txdepthget = fn->txdepthget;
	pioh->pio_fn.txdepthset = fn->txdepthset;

	pioh->pio_fn.tx = fn->tx;
	pioh->pio_fn.rx = fn->rx;
}

static int
_wlc_pio_com_detach(pio_info_t *pt)
{
	/* assert pkts are freed */
	ASSERT(pktq_empty(&pt->txpioq));

	if (pt)
		MFREE(pt->pub->osh, pt, sizeof(pio_info_t));
	return 0;
}

/* pio initialization */
static int
_wlc_pio_com_init(pio_info_t *pt)
{
	return 0;
}

static int
_wlc_pio_com_reset(pio_info_t *pt)
{
	return 0;
}

static int
_wlc_pio_com_txsuspend(pio_info_t *pt)
{
	if (pt->pio4) {
		OR_REG(&pt->p4txregs->fifocontrol, XFC4_SE);
	} else {
		OR_REG(&pt->p2txregs->fifocontrol, XFC_SE);
	}
	return 0;
}

static bool
_wlc_pio_com_txsuspended(pio_info_t *pt)
{
	if (pt->pio4) {
		if ((R_REG(&pt->p4txregs->fifocontrol) & (XFC4_SE | XFC4_SP)) == XFC4_SE)
			return TRUE;
	} else {
		if ((R_REG(&pt->p2txregs->fifocontrol) & (XFC_SE | XFC_SP)) == XFC_SE)
			return TRUE;
	}
	return FALSE;
}

static int
_wlc_pio_com_txresume(pio_info_t *pt)
{
	if (pt->pio4) {
		AND_REG(&pt->p4txregs->fifocontrol, ~XFC4_SE);
	} else {
		AND_REG(&pt->p2txregs->fifocontrol, ~XFC_SE);
	}
	return 0;
}

static bool
_wlc_pio_com_rxfrmrdy(pio_info_t *pt)
{
	if (pt->pio4)
		return (R_REG(&pt->p4rxregs->fifocontrol) & RFC_FR);
	else
		return (R_REG(&pt->p2rxregs->fifocontrol) & RFC_FR);
}

static bool
_wlc_pio_com_txavailable(pio_info_t *pt, uint len, uint nfrags)
{
	if (((pt->txfrmcnt + nfrags) <= D11_MAX_TX_FRMS) &&
	    ((pt->txdatacnt + ((nfrags == 1) ?
		(ROUNDUP(len, 4)) : (len + 3*nfrags))) <= pt->txfifolimit)) {
		if (pt->pio4)
			return (!(R_REG(&pt->p4txregs->fifocontrol) & XFC4_SE));
		else
			return (!(R_REG(&pt->p2txregs->fifocontrol) & XFC_SE));
	}

	return FALSE;
}

static int
_wlc_pio_com_cntupd(pio_info_t *pt, uint len)
{
	PIO_SUB_TX_CNT(pt, len);
	return 0;
}

/* need fifo to avoid touch some rx fifo, which doesn't exist */
static char*
_wlc_pio_com_dump(pio_info_t *pt, char *buf)
{
	/* common software part */
	buf += sprintf(buf, "pio%d: qlen %d txfrmcnt %d txdatacnt %d txfifodepth %d\n",
		pt->_fifo, pktq_len(&pt->txpioq), pt->txfrmcnt, pt->txdatacnt, pt->txfifolimit);

	if (pt->pio4) {
		buf += sprintf(buf, "xmtcontrol 0x%x\n", R_REG(&pt->p4txregs->fifocontrol));
		if (pt->_fifo == RX_FIFO) {
			buf += sprintf(buf, "rcvcontrol 0x%x\n", R_REG(&pt->p4rxregs->fifocontrol));
		}
	} else {
		buf += sprintf(buf, "xmtcontrol 0x%x\n", R_REG(&pt->p2txregs->fifocontrol));
		if (pt->_fifo == RX_FIFO) {
			buf += sprintf(buf, "rcvcontrol 0x%x\n", R_REG(&pt->p2rxregs->fifocontrol));
		}
	}
	return (buf);
}

/* 2-byte mode for corerev < 8. poke a chain of fragment buffers into tx channel fifo */
static int
_wlc_pio2_tx(pio_info_t *pt, void *p0)
{
	pio2regs_t *pioregs = pt->p2txregs;
	void *p;
	osl_t *osh;
	uint totlen;
	uchar *va;
	uint len;

	WL_TRACE(("wl%d: _wlc_pio2_tx\n", pt->pub->unit));

	ASSERT(pioregs);
	osh = pt->pub->osh;
	totlen = pkttotlen(osh, p0);

	/* ASSERT(_wlc_pio_tx_available((pio_t *)pt, totlen, 1)); */

	PIO_ADD_TX_CNT(pt, totlen);

	/* clear frameready */
	W_REG(&pioregs->fifocontrol, XFC_FR);

	for (p = p0; p; p = PKTNEXT(osh, p)) {
		va = PKTDATA(osh, p);
		len = PKTLEN(osh, p);

		/* skip any zero-byte buffers */
		if (len == 0)
			continue;

		_wlc_pio2_xmtfifo(pt, pioregs, va, len);
		totlen -= len;
	}

	W_REG(&pioregs->fifocontrol, XFC_EF);

	/* save frag in pio sw queue */
	pktenq(&pt->txpioq, p0);
	return 0;
}

/* 2-byte mode for corerev < 8.  move tx data to fifo */
static void
_wlc_pio2_xmtfifo(pio_info_t *pt, pio2regs_t *regs, uchar *va, uint len)
{
	uint16 *va16;
	volatile uint16 *fifo;

	if (len == 0)
		return;

	fifo = &regs->fifodata;

	/* write any odd leading byte */
	if ((uintptr)va & 1) {
		W_REG(&regs->fifocontrol, XFC_LO);
		W_REG(fifo, (uint16)*va);
		va++;
		len--;
	}

	va16 = (uint16*)va;

	if (len >= 2) {
		W_REG(&regs->fifocontrol, XFC_BOTH);

		{
			while (len >= 2) {
				W_REG(fifo, *va16);
				va16++;
				len -= 2;
			}
		}
	}

	/* write any odd trailing byte */
	if (len) {
		W_REG(&regs->fifocontrol, XFC_LO);
		W_REG(fifo, (uint16)*(uint8*)va16);
	}
}

/* 4-byte mode for corerev >= 8. poke a chain of fragment buffers into tx channel fifo */
static int
_wlc_pio4_tx(pio_info_t *pt, void *p0)
{
	pio4regs_t *pioregs = pt->p4txregs;
	void *p;
	osl_t *osh;
	uint totlen;
	uchar *va;
	uint len;

	WL_TRACE(("wl%d: _wlc_pio4_tx\n", pt->pub->unit));

	ASSERT(pioregs);
	osh = pt->pub->osh;
	totlen = pkttotlen(osh, p0);

	/* ASSERT(_wlc_pio_tx_available((pio_t *)pt, totlen, 1)); */

	PIO_ADD_TX_CNT(pt, totlen);

	/* clear frameready */
	W_REG(&pioregs->fifocontrol, XFC4_FR);

	for (p = p0; p; p = PKTNEXT(osh, p)) {
		va = PKTDATA(osh, p);
		len = PKTLEN(osh, p);

		/* skip any zero-byte buffers */
		if (len == 0)
			continue;

		_wlc_pio4_xmtfifo(pt, pioregs, va, len);
		totlen -= len;
	}

	W_REG(&pioregs->fifocontrol, XFC4_EF);

	/* save frag in pio sw queue */
	pktenq(&pt->txpioq, p0);
	return 0;
}

/* 4-byte mode for corerev >= 8.  move tx data to fifo */
static void
_wlc_pio4_xmtfifo(pio_info_t *pt, pio4regs_t *regs, uchar *va, uint len)
{
	uint i;
	uint32 *va32;
	volatile uint32 *fifo;
#ifndef IL_BIGENDIAN
	uint byte_valid[] = {0, 1, 3, 7}; /* for first byte, first word, first three bytes */
#else
	uint byte_valid[] = {0, 8, 12, 14}; /* for first byte, first word, first three bytes */
#endif /* IL_BIGENDIAN */

	if (len == 0)
		return;

	fifo = &regs->fifodata;

	/* write any leading bytes which not aligned to 32-bit boundary */
	i = (uint)((uintptr)va & 3);
	if (i) {
		i = MIN(i, len);
		W_REG(&regs->fifocontrol, byte_valid[i]);
		W_REG(fifo, *(uint32*)va);
		va += i;
		len -= i;
	}

	va32 = (uint32*)va;

	if (len >= 4) {
		{
			/* all 4 bytes are valid */
			W_REG(&regs->fifocontrol, 0xf);
			while (len >= 4) {
				W_REG(fifo, *va32);
				va32++;
				len -= 4;
			}
		}
	}

	if (len) {
		/* write any trailing bytes */
		W_REG(&regs->fifocontrol, byte_valid[len]);
		W_REG(fifo, *va32);
	}
}


/* shared rx process between pio2 and pio4 */
static bool
_wlc_pio_rxcheck(pio_info_t *pt, int len, void **p)
{
	/* flush giant frames */
	if (len > RXBUFSZ) {
		WL_ERROR(("wl%d: %s: giant frame. len %d\n", pt->pub->unit, __FUNCTION__, len));
		WLCNTINCR(pt->pub->_cnt.rxgiant);
		return FALSE;
	}
	/* flush runt frames if not tx status frame */
	if (len == 0 && pt->_fifo != RX_TXSTATUS_FIFO) {
		WL_ERROR(("wl%d: %s: runt frame\n", pt->pub->unit, __FUNCTION__));
		WLCNTINCR(pt->pub->_cnt.rxrunt);
		return FALSE;
	}
	/* alloc a new buf-wlc_recvdata() fragment reassembly requires RXBUFSZ receive buffers */
	if ((*p = PKTGET(pt->pub->osh, RXBUFSZ, FALSE)) == NULL) {
		WL_ERROR(("wl%d: %s: out of rxbufs\n", pt->pub->unit, __FUNCTION__));
		WLCNTINCR(pt->pub->_cnt.rxnobuf);
		return FALSE;
	}

	return TRUE;
}

/* 2-byte mode for corerev < 8.  move rx fifo data to buffer */
static void*
_wlc_pio2_rx(pio_info_t *pt)
{
	void *p = NULL;
	osl_t *osh;
	uint8 *va;
	uint16 rxstatus;
	uint32 fifocontrol;
	d11rxhdr_t *rxh;
	int i, len, rxhdrlen = 0;
	pio2regs_t *regs;

	uint16 *va16;
	volatile uint16 *dptr, w;

	WL_TRACE(("wl%d: _wlc_pio2_rx\n", pt->pub->unit));

	osh = pt->pub->osh;
	regs = pt->p2rxregs;
	ASSERT(regs);

	/* if frame is not ready, return */
	if (!(R_REG(&regs->fifocontrol) & RFC_FR))
		return NULL;
	/* clear frame ready */
	W_REG(&regs->fifocontrol, RFC_FR);
	/* wait for data valid */
	SPINWAIT((((fifocontrol = R_REG(&regs->fifocontrol)) & RFC_DR) == 0), 100);
	if ((fifocontrol & RFC_DR) == 0) {
		WL_ERROR(("wl%d: %s: data ready never set\n", pt->pub->unit, __FUNCTION__));
		return NULL;
	}

	dptr = &regs->fifodata;	/* rcv data */
	/* read the first 16bits of the rxhdr which gives the length */
	w = R_REG(dptr);
	/* hanlde both Big and Little endian */
	len = ltoh16(w);

	if (!_wlc_pio_rxcheck(pt, len, &p))
		goto flush;

	va = (uint8 *)PKTDATA(osh, p);
	ASSERT(ISALIGNED((uintptr)va, sizeof(uint32)));
	PKTSETLEN(osh, p, (HWRXOFF + len));
	/* get header length */
	rxhdrlen = (pt->_fifo == RX_TXSTATUS_FIFO) ? TXSTATUS_LEN : RXHDR_LEN;


	/* read the rxheader from the fifo first */
	va16 = (uint16 *)va;
	*va16++ = w;
	i = (rxhdrlen / 2) - 1;
	{
		while (i--) {
			*va16 = R_REG(dptr);
			va16++;
		}
	}

	/* toss bad packet before read/pass up whole frame */
	rxh = (d11rxhdr_t*)va;
	rxstatus = ltoh16(rxh->RxStatus1);
	if (rxstatus & RXS_FCSERR) {
		/* ucode should toss it */
		ASSERT(0);
		goto flush;
	}

	/* tx status frame only has header */
	if (pt->_fifo == RX_TXSTATUS_FIFO)
		return (p);

	/* read the frame */
	va16 = (uint16*) &va[HWRXOFF];

	if (rxstatus & RXS_PBPRES) {
		/* advance data pointer by two-byte for address alignment */
		va16++;
		PKTSETLEN(osh, p, (HWRXOFF + len + 2));
	}

	{
		i = len / 2;
		while (i--) {
			*va16 = R_REG(dptr);
			va16++;
		}
		/* read last (odd) byte */
		if (len & 1) {
			*va16 = R_REG(dptr) & 0xff;
		}
	}


	return (p);

flush:
	/* flush rx frame */
	W_REG(&regs->fifocontrol, RFC_DR);
	if (p)
		PKTFREE(osh, p, FALSE);

	return (NULL);
}

/* 4-byte mode for corerev >= 8.  move rx fifo data to buffer */
static void*
_wlc_pio4_rx(pio_info_t *pt)
{
	void *p = NULL;
	osl_t *osh;
	uint8 *va;
	uint16 rxstatus;
	uint32 fifocontrol;
	d11rxhdr_t *rxh;
	int i, len, rxhdrlen = 0;
	pio4regs_t *regs;
	uint32 *va32;
	volatile uint32 *dptr, w;

	WL_TRACE(("wl%d: _wlc_pio4_rx\n", pt->pub->unit));

	osh = pt->pub->osh;
	regs = pt->p4rxregs;
	ASSERT(regs);

	/* if frame is not ready, return */
	if (!(R_REG(&regs->fifocontrol) & RFC_FR))
		return NULL;
	/* clear frame ready */
	W_REG(&regs->fifocontrol, RFC_FR);
	/* wait for data valid */
	SPINWAIT((((fifocontrol = R_REG(&regs->fifocontrol)) & RFC_DR) == 0), 100);
	if ((fifocontrol & RFC_DR) == 0) {
		WL_ERROR(("wl%d: %s: data ready never set\n", pt->pub->unit, __FUNCTION__));
		return NULL;
	}

	dptr = &regs->fifodata;	/* rcv data */
	/* read the first 16bits of the rxhdr which gives the length */
	w = R_REG(dptr);
	/* hanlde both Big and Little endian */
	len = (uint16)ltoh32(w);

	if (!_wlc_pio_rxcheck(pt, len, &p))
		goto flush;

	va = (uint8 *)PKTDATA(osh, p);
	ASSERT(ISALIGNED((uintptr)va, sizeof(uint32)));
	PKTSETLEN(osh, p, (HWRXOFF + len));
	/* get header length */
	rxhdrlen = (pt->_fifo == RX_TXSTATUS_FIFO) ? TXSTATUS_LEN : RXHDR_LEN;


	/* read the rxheader from the fifo first */
	va32 = (uint32 *)va;
	*va32++ = w;
	i = (rxhdrlen / 4) - 1;
	{
		while (i--) {
			*va32 = R_REG(dptr);
			va32++;
		}
	}

	if (rxhdrlen % 4) {
		ASSERT((rxhdrlen % 4) == 2);
		*((uint16 *)va32) = R_REG((volatile uint16 *)dptr);
		va32 = (uint32 *)((int8*)va32 + 2);
	}

	/* toss bad packet before read/pass up whole frame */
	rxh = (d11rxhdr_t*)va;
	rxstatus = ltoh16(rxh->RxStatus1);
	if (rxstatus & RXS_FCSERR) {
		/* ucode should toss it */
		ASSERT(0);
		goto flush;
	}

	/* tx status frame only has header */
	if (pt->_fifo == RX_TXSTATUS_FIFO)
		return (p);

	/* read the frame */
	va32 = (uint32*) &va[HWRXOFF];

	/* align buffer to 32-bit boundary */
	if ((uintptr)va32 % 4 && len >= 2) {
		*((uint16 *)va32) = R_REG((volatile uint16 *)dptr);
		va32 = (uint32 *)((int8*)va32 + 2);
		len -= 2;
	}

	i = len / 4;

	{
		while (i--) {
			*va32 = R_REG(dptr);
			va32++;
		}
	}

	/* read the remain bytes if exist */
	i = len % 4;
	if (i) {
		uint32 mask[] = {0, 0x000000ff, 0x0000ffff, 0x00ffffff};
		*va32 = R_REG(dptr) & mask[i];
	}


	return (p);

flush:
	/* flush rx frame */
	W_REG(&regs->fifocontrol, RFC_DR);
	if (p)
		PKTFREE(osh, p, FALSE);

	return (NULL);
}

static void*
_wlc_pio_getnexttxp(pio_info_t *pt)
{
	return pktdeq(&pt->txpioq);
}

static int
_wlc_pio_txreclaim(pio_info_t *pt)
{
	void *p;
	while ((p = pktdeq(&pt->txpioq)))
		PKTFREE(pt->pub->osh, p, TRUE);

	/* reset per-fifo software counters */
	pt->txfrmcnt = 0;
	pt->txdatacnt = 0;
	return 0;
}

static int
_wlc_pio_txfifodepthset(pio_info_t *pt, uint len)
{
	pt->txfifolimit = len;
	return 0;
}

static uint
_wlc_pio_txfifodepthget(pio_info_t *pt)
{
	return pt->txfifolimit;
}
